@@ -17,6 +17,8 @@
 
 """An adapter between a Git Branch and a Bazaar Branch"""
 
+from collections import defaultdict
+
 from dulwich.objects import (
     Commit,
     Tag,
@@ -51,9 +53,12 @@ from bzrlib.plugins.git.errors import (
     NoSuchRef,
     )
 from bzrlib.plugins.git.refs import (
-    ref_to_branch_name,
     extract_tags,
+    is_tag,
+    ref_to_branch_name,
+    ref_to_tag_name,
     tag_name_to_ref,
+    UnpeelMap,
     )
 
 from bzrlib.foreign import ForeignBranch
@@ -76,48 +81,135 @@ class GitPullResult(branch.PullResult):
         return self._lookup_revno(self.new_revid)
 
 
-class LocalGitTagDict(tag.BasicTags):
-    """Dictionary with tags in a local repository."""
+class GitTags(tag.BasicTags):
+    """Ref-based tag dictionary."""
 
     def __init__(self, branch):
         self.branch = branch
         self.repository = branch.repository
 
+    def get_refs(self):
+        raise NotImplementedError(self.get_refs)
+
+    def _iter_tag_refs(self, refs):
+        raise NotImplementedError(self._iter_tag_refs)
+
+    def _merge_to_git(self, to_tags, refs, overwrite=False):
+        target_repo = to_tags.repository
+        conflicts = []
+        for k, v in refs.iteritems():
+            if not is_tag(k):
+                continue
+            if overwrite or not k in self.target.repository.refs:
+                target_repo.refs[k] = v
+            elif target_repo.repository.refs[k] == v:
+                pass
+            else:
+                conflicts.append((ref_to_tag_name(k), v, target_repo.refs[k]))
+        return conflicts
+
+    def _merge_to_non_git(self, to_tags, refs, overwrite=False):
+        unpeeled_map = defaultdict(set)
+        conflicts = []
+        result = dict(to_tags.get_tag_dict())
+        for n, peeled, unpeeled, bzr_revid in self._iter_tag_refs(refs):
+            if unpeeled is not None:
+                unpeeled_map[peeled].add(unpeeled)
+            if n not in result or overwrite:
+                result[n] = bzr_revid
+            elif result[n] == bzr_revid:
+                pass
+            else:
+                conflicts.append((n, result[n], bzr_revid))
+        to_tags._set_tag_dict(result)
+        if len(unpeeled_map) > 0:
+            map_file = UnpeelMap.from_repository(to_tags.branch.repository)
+            map_file.update(unpeeled_map)
+            map_file.save_in_repository(to_tags.branch.repository)
+        return conflicts
+
+    def merge_to(self, to_tags, overwrite=False, ignore_master=False,
+                 source_refs=None):
+        if source_refs is None:
+            source_refs = self.get_refs()
+        if self == to_tags:
+            return
+        if isinstance(to_tags, GitTags):
+            return self._merge_to_git(to_tags, source_refs,
+                                      overwrite=overwrite)
+        else:
+            if ignore_master:
+                master = None
+            else:
+                master = to_tags.branch.get_master_branch()
+            conflicts = self._merge_to_non_git(to_tags, source_refs,
+                                              overwrite=overwrite)
+            if master is not None:
+                conflicts += self.merge_to(to_tags, overwrite=overwrite,
+                                           source_refs=source_refs,
+                                           ignore_master=ignore_master)
+            return conflicts
+
     def get_tag_dict(self):
         ret = {}
-        for k,v in extract_tags(self.repository._git.get_refs()).iteritems():
+        refs = self.get_refs()
+        for (name, peeled, unpeeled, bzr_revid) in self._iter_tag_refs(refs):
+            ret[name] = bzr_revid
+        return ret
+
+
+class LocalGitTagDict(GitTags):
+    """Dictionary with tags in a local repository."""
+
+    def __init__(self, branch):
+        super(LocalGitTagDict, self).__init__(branch)
+        self.refs = self.repository._git.refs
+
+    def get_refs(self):
+        return self.repository._git.get_refs()
+
+    def _iter_tag_refs(self, refs):
+        """Iterate over the tag refs.
+
+        :param refs: Refs dictionary (name -> git sha1)
+        :return: iterator over (name, peeled_sha1, unpeeled_sha1, bzr_revid)
+        """
+        for k, (peeled, unpeeled) in extract_tags(refs).iteritems():
             try:
-                obj = self.repository._git[v]
+                obj = self.repository._git[peeled]
             except KeyError:
-                mutter("Tag %s points at unknown object %s, ignoring", v, obj)
+                mutter("Tag %s points at unknown object %s, ignoring", peeled,
+                       obj)
                 continue
+            # FIXME: this shouldn't really be necessary, the repository
+            # already should have these unpeeled.
             while isinstance(obj, Tag):
-                v = obj.object[1]
-                obj = self.repository._git[v]
+                peeled = obj.object[1]
+                obj = self.repository._git[peeled]
             if not isinstance(obj, Commit):
                 mutter("Tag %s points at object %r that is not a commit, "
                        "ignoring", k, obj)
                 continue
-            ret[k] = self.branch.lookup_foreign_revision_id(v)
-        return ret
+            yield (k, peeled, unpeeled,
+                   self.branch.lookup_foreign_revision_id(peeled))
 
     def _set_tag_dict(self, to_dict):
-        extra = set(self.repository._git.get_refs().keys())
+        extra = set(self.get_refs().keys())
         for k, revid in to_dict.iteritems():
             name = tag_name_to_ref(k)
             if name in extra:
                 extra.remove(name)
             self.set_tag(k, revid)
         for name in extra:
-            if name.startswith("refs/tags/"):
+            if is_tag(name):
                 del self.repository._git[name]
 
     def set_tag(self, name, revid):
-        self.repository._git.refs[tag_name_to_ref(name)], _ = \
-            self.branch.mapping.revision_id_bzr_to_foreign(revid)
+        self.refs[tag_name_to_ref(name)], _ = \
+            self.branch.lookup_bzr_revision_id(revid)
 
 
-class DictTagDict(LocalGitTagDict):
+class DictTagDict(tag.BasicTags):
 
     def __init__(self, branch, tags):
         super(DictTagDict, self).__init__(branch)
@@ -257,6 +349,10 @@ class GitBranch(ForeignBranch):
         return self.repository.lookup_foreign_revision_id(foreign_revid,
             self.mapping)
 
+    def lookup_bzr_revision_id(self, revid):
+        return self.repository.lookup_bzr_revision_id(
+            revid, mapping=self.mapping)
+
 
 class LocalGitBranch(GitBranch):
     """A local Git branch."""
@@ -347,6 +443,20 @@ class LocalGitBranch(GitBranch):
         return True
 
 
+def _quick_lookup_revno(local_branch, remote_branch, revid):
+    assert isinstance(revid, str), "was %r" % revid
+    # Try in source branch first, it'll be faster
+    try:
+        return local_branch.revision_id_to_revno(revid)
+    except errors.NoSuchRevision:
+        graph = local_branch.repository.get_graph()
+        try:
+            return graph.find_distance_to_null(revid)
+        except errors.GhostRevisionsHaveNoRevno:
+            # FIXME: Check using graph.find_distance_to_null() ?
+            return remote_branch.revision_id_to_revno(revid)
+
+
 class GitBranchPullResult(branch.PullResult):
 
     def __init__(self):
@@ -367,13 +477,7 @@ class GitBranchPullResult(branch.PullResult):
         self._show_tag_conficts(to_file)
 
     def _lookup_revno(self, revid):
-        assert isinstance(revid, str), "was %r" % revid
-        # Try in source branch first, it'll be faster
-        try:
-            return self.source_branch.revision_id_to_revno(revid)
-        except errors.NoSuchRevision:
-            # FIXME: Check using graph.find_distance_to_null() ?
-            return self.target_branch.revision_id_to_revno(revid)
+        return _quick_lookup_revno(self.target_branch, self.source_branch, revid)
 
     def _get_old_revno(self):
         if self._old_revno is not None:
@@ -399,13 +503,7 @@ class GitBranchPullResult(branch.PullResult):
 class GitBranchPushResult(branch.BranchPushResult):
 
     def _lookup_revno(self, revid):
-        assert isinstance(revid, str), "was %r" % revid
-        # Try in source branch first, it'll be faster
-        try:
-            return self.source_branch.revision_id_to_revno(revid)
-        except errors.NoSuchRevision:
-            # FIXME: Check using graph.find_distance_to_null() ?
-            return self.target_branch.revision_id_to_revno(revid)
+        return _quick_lookup_revno(self.source_branch, self.target_branch, revid)
 
     @property
     def old_revno(self):
@@ -413,12 +511,21 @@ class GitBranchPushResult(branch.BranchPushResult):
 
     @property
     def new_revno(self):
+        new_original_revno = getattr(self, "new_original_revno", None)
+        if new_original_revno:
+            return new_original_revno
+        if getattr(self, "new_original_revid", None) is not None:
+            return self._lookup_revno(self.new_original_revid)
         return self._lookup_revno(self.new_revid)
 
 
 class InterFromGitBranch(branch.GenericInterBranch):
     """InterBranch implementation that pulls from Git into bzr."""
 
+    @staticmethod
+    def _get_branch_formats_to_test():
+        return []
+
     @classmethod
     def _get_interrepo(self, source, target):
         return repository.InterRepository.get(source.repository,
@@ -436,7 +543,7 @@ class InterFromGitBranch(branch.GenericInterBranch):
 
         Compared to the `update_revisions()` below, this function takes a
         `limit` argument that limits how many git commits will be converted
-        and returns the new git head.
+        and returns the new git head and remote refs.
         """
         interrepo = self._get_interrepo(self.source, self.target)
         def determine_wants(heads):
@@ -455,7 +562,7 @@ class InterFromGitBranch(branch.GenericInterBranch):
             if self.target.repository.has_revision(self._last_revid):
                 return []
             return [head]
-        pack_hint, head = interrepo.fetch_objects(
+        pack_hint, head, refs = interrepo.fetch_objects(
             determine_wants, self.source.mapping, limit=limit)
         if (pack_hint is not None and
             self.target.repository._format.pack_compresses):
@@ -468,7 +575,7 @@ class InterFromGitBranch(branch.GenericInterBranch):
             prev_last_revid = self.target.last_revision()
         self.target.generate_revision_history(self._last_revid,
             prev_last_revid)
-        return head
+        return head, refs
 
     def update_revisions(self, stop_revision=None, overwrite=False,
                          graph=None):
@@ -506,7 +613,7 @@ class InterFromGitBranch(branch.GenericInterBranch):
             graph = self.target.repository.get_graph(self.source.repository)
             (result.old_revno, result.old_revid) = \
                 self.target.last_revision_info()
-            result.new_git_head = self._update_revisions(
+            result.new_git_head, remote_refs = self._update_revisions(
                 stop_revision, overwrite=overwrite, graph=graph, limit=limit)
             result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
                 overwrite)
@@ -531,7 +638,7 @@ class InterFromGitBranch(branch.GenericInterBranch):
         result.target_branch = self.target
         graph = self.target.repository.get_graph(self.source.repository)
         result.old_revno, result.old_revid = self.target.last_revision_info()
-        result.new_git_head = self._update_revisions(
+        result.new_git_head, remote_refs = self._update_revisions(
             stop_revision, overwrite=overwrite, graph=graph)
         result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
             overwrite)
@@ -546,6 +653,10 @@ class InterGitBranch(branch.GenericInterBranch):
 class InterGitLocalRemoteBranch(InterGitBranch):
     """InterBranch that copies from a local to a remote git branch."""
 
+    @staticmethod
+    def _get_branch_formats_to_test():
+        return []
+
     @classmethod
     def is_compatible(self, source, target):
         from bzrlib.plugins.git.remote import RemoteGitBranch
@@ -575,6 +686,10 @@ class InterGitLocalRemoteBranch(InterGitBranch):
 class InterGitRemoteLocalBranch(InterGitBranch):
     """InterBranch that copies from a remote to a local git branch."""
 
+    @staticmethod
+    def _get_branch_formats_to_test():
+        return []
+
     @classmethod
     def is_compatible(self, source, target):
         from bzrlib.plugins.git.remote import RemoteGitBranch
@@ -588,15 +703,11 @@ class InterGitRemoteLocalBranch(InterGitBranch):
         result.old_revid = self.target.last_revision()
         refs, stop_revision = self.update_refs(stop_revision)
         self.target.generate_revision_history(stop_revision, result.old_revid)
-        self.update_tags(refs)
+        result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
+            source_refs=refs, overwrite=overwrite)
         result.new_revid = self.target.last_revision()
         return result
 
-    def update_tags(self, refs):
-        for name, v in extract_tags(refs).iteritems():
-            revid = self.target.lookup_foreign_revision_id(v)
-            self.target.tags.set_tag(name, revid)
-
     def update_refs(self, stop_revision=None):
         interrepo = repository.InterRepository.get(self.source.repository,
             self.target.repository)
@@ -618,7 +729,8 @@ class InterGitRemoteLocalBranch(InterGitBranch):
         result.old_revid = self.target.last_revision()
         refs, stop_revision = self.update_refs(stop_revision)
         self.target.generate_revision_history(stop_revision, result.old_revid)
-        self.update_tags(refs)
+        result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
+            overwrite=overwrite, source_refs=refs)
         result.new_revid = self.target.last_revision()
         return result
 
@@ -645,14 +757,14 @@ class InterToGitBranch(branch.GenericInterBranch):
 
     def _get_new_refs(self, stop_revision=None):
         if stop_revision is None:
-            stop_revision = self.source.last_revision()
+            (stop_revno, stop_revision) = self.source.last_revision_info()
         assert type(stop_revision) is str
         main_ref = self.target.ref or "refs/heads/master"
         refs = { main_ref: (None, stop_revision) }
         for name, revid in self.source.tags.get_tag_dict().iteritems():
             if self.source.repository.has_revision(revid):
                 refs[tag_name_to_ref(name)] = (None, revid)
-        return refs, main_ref
+        return refs, main_ref, (stop_revno, stop_revision)
 
     def pull(self, overwrite=False, stop_revision=None, local=False,
              possible_transports=None):
@@ -660,16 +772,17 @@ class InterToGitBranch(branch.GenericInterBranch):
         result = GitBranchPullResult()
         result.source_branch = self.source
         result.target_branch = self.target
-        new_refs, main_ref = self._get_new_refs(stop_revision)
+        new_refs, main_ref, stop_revinfo = self._get_new_refs(stop_revision)
         def update_refs(old_refs):
             refs = dict(old_refs)
             # FIXME: Check for diverged branches
             refs.update(new_refs)
             return refs
         old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
-        result.old_revid = self.target.lookup_foreign_revision_id(
-            old_refs.get(main_ref, ZERO_SHA))
-        result.new_revid = new_refs[main_ref]
+        (result.old_revid, old_sha1) = old_refs.get(main_ref, (ZERO_SHA, NULL_REVISION))
+        if result.old_revid is None:
+            result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
+        result.new_revid = new_refs[main_ref][1]
         return result
 
     def push(self, overwrite=False, stop_revision=None,
@@ -678,23 +791,24 @@ class InterToGitBranch(branch.GenericInterBranch):
         result = GitBranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
-        new_refs, main_ref = self._get_new_refs(stop_revision)
+        new_refs, main_ref, stop_revinfo = self._get_new_refs(stop_revision)
         def update_refs(old_refs):
             refs = dict(old_refs)
             # FIXME: Check for diverged branches
             refs.update(new_refs)
             return refs
         old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
-        result.old_revid = self.target.lookup_foreign_revision_id(
-            old_refs.get(main_ref, ZERO_SHA))
-        result.new_revid = new_refs[main_ref]
+        (result.old_revid, old_sha1) = old_refs.get(main_ref, (ZERO_SHA, NULL_REVISION))
+        if result.old_revid is None:
+            result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
+        result.new_revid = new_refs[main_ref][1]
         return result
 
     def lossy_push(self, stop_revision=None):
         result = GitBranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
-        new_refs, main_ref = self._get_new_refs(stop_revision)
+        new_refs, main_ref, stop_revinfo = self._get_new_refs(stop_revision)
         def update_refs(old_refs):
             refs = dict(old_refs)
             # FIXME: Check for diverged branches
@@ -704,6 +818,7 @@ class InterToGitBranch(branch.GenericInterBranch):
             update_refs)
         result.old_revid = old_refs.get(self.target.ref, (None, NULL_REVISION))[1]
         result.new_revid = new_refs[main_ref][1]
+        (result.new_original_revno, result.new_original_revid) = stop_revinfo
         return result
 
 
