@@ -37,6 +37,7 @@ class cmd_git_import(Command):
     takes_args = ["src_location", "dest_location?"]
 
     def run(self, src_location, dest_location=None):
+        from collections import defaultdict
         import os
         from bzrlib import (
             ui,
@@ -58,6 +59,7 @@ class cmd_git_import(Command):
             GitBranch,
             extract_tags,
             )
+        from bzrlib.plugins.git.refs import ref_to_branch_name
         from bzrlib.plugins.git.repository import GitRepository
 
         if dest_location is None:
@@ -80,14 +82,21 @@ class cmd_git_import(Command):
 
         interrepo = InterRepository.get(source_repo, target_repo)
         mapping = source_repo.get_mapping()
-        refs = interrepo.fetch_refs()
+        refs = interrepo.fetch()
+        unpeeled_tags = defaultdict(set)
         tags = {}
-        for k, v in extract_tags(refs).iteritems():
-            tags[k] = mapping.revision_id_foreign_to_bzr(v)
+        for k, (peeled, unpeeled) in extract_tags(refs).iteritems():
+            tags[k] = mapping.revision_id_foreign_to_bzr(peeled)
+            if unpeeled is not None:
+                unpeeled_tags[peeled].add(unpeeled)
+        # FIXME: Store unpeeled tag map
         pb = ui.ui_factory.nested_progress_bar()
         try:
             for i, (name, ref) in enumerate(refs.iteritems()):
-                if name.startswith("refs/tags/"):
+                try:
+                    ref_to_branch_name(name)
+                except ValueError:
+                    # Not a branch, ignore
                     continue
                 pb.update("creating branches", i, len(refs))
                 head_loc = os.path.join(dest_location, name)
@@ -97,7 +106,7 @@ class cmd_git_import(Command):
                     parent_path = urlutils.dirname(head_loc)
                     if not os.path.isdir(parent_path):
                         os.makedirs(parent_path)
-                    head_bzrdir = BzrDir.create(head_loc, format=format)
+                    head_bzrdir = BzrDir.create(head_loc)
                 try:
                     head_branch = head_bzrdir.open_branch()
                 except NotBranchError:
@@ -204,27 +213,49 @@ class cmd_git_apply(Command):
     "bzr pull".
     """
 
+    takes_options = [
+        Option('signoff', short_name='s', help='Add a Signed-off-by line.'),
+        'force']
     takes_args = ["patches*"]
 
-    def _apply_patch(self, wt, f):
-        from dulwich.patch import git_am_patch_split
-        (c, diff, version) = git_am_patch_split(f)
-        # FIXME: Process diff
-        wt.commit(committer=c.committer,
-                  message=c.message)
+    def _apply_patch(self, wt, f, signoff):
+        """Apply a patch.
 
-    def run(self, patches_list=None):
+        :param wt: A Bazaar working tree object.
+        :param f: Patch file to read.
+        :param signoff: Add Signed-Off-By flag.
+        """
+        from bzrlib.errors import BzrCommandError
+        from dulwich.patch import git_am_patch_split
+        import subprocess
+        (c, diff, version) = git_am_patch_split(f)
+        # FIXME: Cope with git-specific bits in patch
+        p = subprocess.Popen(["patch", "-p1"], stdin=subprocess.PIPE, cwd=wt.basedir)
+        p.communicate(diff)
+        exitcode = p.wait()
+        if exitcode != 0:
+            raise BzrCommandError("error running patch")
+        message = c.message
+        if signoff:
+            signed_off_by = wt.branch.get_config().username()
+            message += "Signed-off-by: %s\n" % signed_off_by.encode('utf-8')
+        wt.commit(authors=[c.author], message=message)
+
+    def run(self, patches_list=None, signoff=False, force=False):
+        from bzrlib.errors import UncommittedChanges
         from bzrlib.workingtree import WorkingTree
         if patches_list is None:
             patches_list = []
-        
+
         tree, _ = WorkingTree.open_containing(".")
+        if tree.basis_tree().changes_from(tree).has_changed() and not force:
+            raise UncommittedChanges(tree)
         tree.lock_write()
         try:
             for patch in patches_list:
                 f = open(patch, 'r')
                 try:
-                    self._apply_patch(tree, f)
+                    self._apply_patch(tree, f, signoff=signoff)
                 finally:
                     f.close()
         finally:
