@@ -15,10 +15,9 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-"""An adapter between a Git control dir and a Bazaar BzrDir."""
+"""An adapter between a Git control dir and a Bazaar ControlDir."""
 
 from bzrlib import (
-    bzrdir,
     errors as bzr_errors,
     lockable_files,
     urlutils,
@@ -28,8 +27,21 @@ from bzrlib import (
 LockWarner = getattr(lockable_files, "_LockWarner", None)
 
 from bzrlib.plugins.git import (
-    LocalGitBzrDirFormat,
+    BareLocalGitControlDirFormat,
+    LocalGitControlDirFormat,
     )
+try:
+    from bzrlib.controldir import (
+        ControlDir,
+        format_registry,
+        )
+except ImportError:
+    # bzr < 2.3
+    from bzrlib.bzrdir import (
+        BzrDir,
+        format_registry,
+        )
+    ControlDir = BzrDir
 
 
 class GitLock(object):
@@ -69,21 +81,36 @@ class GitLockableFiles(lockable_files.LockableFiles):
             self._lock_warner = LockWarner(repr(self))
 
 
-class GitDir(bzrdir.BzrDir):
+class GitDirConfig(object):
+
+    def get_default_stack_on(self):
+        return None
+
+    def set_default_stack_on(self, value):
+        raise bzr_errors.BzrError("Cannot set configuration")
+
+
+class GitDir(ControlDir):
     """An adapter to the '.git' dir used by git."""
 
     def is_supported(self):
         return True
 
+    def can_convert_format(self):
+        return False
+
+    def break_lock(self):
+        pass
+
     def cloning_metadir(self, stacked=False):
-        return bzrdir.format_registry.make_bzrdir("default")
+        return format_registry.make_bzrdir("default")
 
     def _branch_name_to_ref(self, name):
         raise NotImplementedError(self._branch_name_to_ref)
 
     if bzrlib_version >= (2, 2):
-        def open_branch(self, name=None, ignore_fallbacks=None,
-            unsupported=False):
+        def open_branch(self, name=None, unsupported=False, 
+            ignore_fallbacks=None):
             return self._open_branch(name=name,
                 ignore_fallbacks=ignore_fallbacks, unsupported=unsupported)
     else:
@@ -91,6 +118,9 @@ class GitDir(bzrdir.BzrDir):
             return self._open_branch(name=None,
                 ignore_fallbacks=ignore_fallbacks, unsupported=unsupported)
 
+    def get_config(self):
+        return GitDirConfig()
+
 
 class LocalGitDir(GitDir):
     """An adapter to the '.git' dir used by git."""
@@ -101,9 +131,18 @@ class LocalGitDir(GitDir):
 
     _gitrepository_class = property(_get_gitrepository_class)
 
+    @property
+    def user_transport(self):
+        return self.root_transport
+
+    @property
+    def control_transport(self):
+        return self.transport
+
     def __init__(self, transport, lockfiles, gitrepo, format):
         self._format = format
         self.root_transport = transport
+        self._mode_check_done = False
         self._git = gitrepo
         if gitrepo.bare:
             self.transport = transport
@@ -125,24 +164,24 @@ class LocalGitDir(GitDir):
     def is_control_filename(self, filename):
         return filename == '.git' or filename.startswith('.git/')
 
-    def get_branch_transport(self, branch_format, name):
+    def get_branch_transport(self, branch_format, name=None):
         if branch_format is None:
             return self.transport
-        if isinstance(branch_format, LocalGitBzrDirFormat):
+        if isinstance(branch_format, LocalGitControlDirFormat):
             return self.transport
         raise bzr_errors.IncompatibleFormat(branch_format, self._format)
 
     def get_repository_transport(self, format):
         if format is None:
             return self.transport
-        if isinstance(format, LocalGitBzrDirFormat):
+        if isinstance(format, LocalGitControlDirFormat):
             return self.transport
         raise bzr_errors.IncompatibleFormat(format, self._format)
 
     def get_workingtree_transport(self, format):
         if format is None:
             return self.transport
-        if isinstance(format, LocalGitBzrDirFormat):
+        if isinstance(format, LocalGitControlDirFormat):
             return self.transport
         raise bzr_errors.IncompatibleFormat(format, self._format)
 
@@ -154,7 +193,20 @@ class LocalGitDir(GitDir):
             self._lockfiles)
 
     def destroy_branch(self, name=None):
-        del self._git.refs[self._branch_name_to_ref(name)]
+        refname = self._branch_name_to_ref(name)
+        if not refname in self._git.refs:
+            raise bzr_errors.NotBranchError(self.root_transport.base,
+                    bzrdir=self)
+        del self._git.refs[refname]
+
+    def destroy_repository(self):
+        raise bzr_errors.UnsupportedOperation(self.destroy_repository, self)
+
+    def destroy_workingtree(self):
+        raise bzr_errors.UnsupportedOperation(self.destroy_workingtree, self)
+
+    def needs_format_conversion(self, format=None):
+        return not isinstance(self._format, format.__class__)
 
     def list_branches(self):
         ret = []
@@ -206,7 +258,7 @@ class LocalGitDir(GitDir):
     def create_workingtree(self, revision_id=None, from_branch=None,
         accelerator_tree=None, hardlink=False):
         if self._git.bare:
-            raise bzr_errors.BzrError("Can't create working tree in a bare repo")
+            raise bzr_errors.UnsupportedOperation(self.create_workingtree, self)
         from dulwich.index import write_index
         from dulwich.pack import SHA1Writer
         f = open(self.transport.local_abspath("index"), 'w+')
@@ -216,3 +268,58 @@ class LocalGitDir(GitDir):
         finally:
             f.close()
         return self.open_workingtree()
+
+    def find_repository(self):
+        """Find the repository that should be used.
+
+        This does not require a branch as we use it to find the repo for
+        new branches as well as to hook existing branches up to their
+        repository.
+        """
+        return self.open_repository()
+
+    def _find_creation_modes(self):
+        """Determine the appropriate modes for files and directories.
+
+        They're always set to be consistent with the base directory,
+        assuming that this transport allows setting modes.
+        """
+        # TODO: Do we need or want an option (maybe a config setting) to turn
+        # this off or override it for particular locations? -- mbp 20080512
+        if self._mode_check_done:
+            return
+        self._mode_check_done = True
+        try:
+            st = self.transport.stat('.')
+        except TransportNotPossible:
+            self._dir_mode = None
+            self._file_mode = None
+        else:
+            # Check the directory mode, but also make sure the created
+            # directories and files are read-write for this user. This is
+            # mostly a workaround for filesystems which lie about being able to
+            # write to a directory (cygwin & win32)
+            if (st.st_mode & 07777 == 00000):
+                # FTP allows stat but does not return dir/file modes
+                self._dir_mode = None
+                self._file_mode = None
+            else:
+                self._dir_mode = (st.st_mode & 07777) | 00700
+                # Remove the sticky and execute bits for files
+                self._file_mode = self._dir_mode & ~07111
+
+    def _get_file_mode(self):
+        """Return Unix mode for newly created files, or None.
+        """
+        if not self._mode_check_done:
+            self._find_creation_modes()
+        return self._file_mode
+
+    def _get_dir_mode(self):
+        """Return Unix mode for newly created directories, or None.
+        """
+        if not self._mode_check_done:
+            self._find_creation_modes()
+        return self._dir_mode
+
+
