@@ -56,6 +56,9 @@ from bzrlib.revision import (
 from bzrlib.revisiontree import (
     RevisionTree,
     )
+from bzrlib.testament import (
+    StrictTestament3,
+    )
 from bzrlib.tsort import (
     topo_sort,
     )
@@ -106,12 +109,14 @@ def import_git_blob(texts, mapping, path, name, (base_hexsha, hexsha),
     else:
         cls = InventoryFile
     ie = cls(file_id, name.decode("utf-8"), parent_id)
-    ie.executable = mode_is_executable(mode)
+    if ie.kind == "file":
+        ie.executable = mode_is_executable(mode)
     if base_hexsha == hexsha and mode_kind(base_mode) == mode_kind(mode):
         base_ie = base_inv[base_inv.path2id(path)]
         ie.text_size = base_ie.text_size
         ie.text_sha1 = base_ie.text_sha1
-        ie.symlink_target = base_ie.symlink_target
+        if ie.kind == "symlink":
+            ie.symlink_target = base_ie.symlink_target
         if ie.executable == base_ie.executable:
             ie.revision = base_ie.revision
         else:
@@ -121,8 +126,6 @@ def import_git_blob(texts, mapping, path, name, (base_hexsha, hexsha),
         if ie.kind == "symlink":
             ie.revision = None
             ie.symlink_target = blob.data
-            ie.text_size = None
-            ie.text_sha1 = None
         else:
             ie.text_size = sum(imap(len, blob.chunked))
             ie.text_sha1 = osutils.sha_strings(blob.chunked)
@@ -155,13 +158,14 @@ def import_git_blob(texts, mapping, path, name, (base_hexsha, hexsha),
                 tuple(parent_keys), ie.text_sha1, chunks)])
     invdelta = []
     if base_hexsha is not None:
-        old_path = path # Renames are not supported yet
+        old_path = path.decode("utf-8") # Renames are not supported yet
         if stat.S_ISDIR(base_mode):
             invdelta.extend(remove_disappeared_children(base_inv, old_path,
                 lookup_object(base_hexsha), [], lookup_object))
     else:
         old_path = None
-    invdelta.append((old_path, path, file_id, ie))
+    new_path = path.decode("utf-8")
+    invdelta.append((old_path, new_path, file_id, ie))
     if base_hexsha != hexsha:
         store_updater.add_object(blob, ie, path)
     return invdelta
@@ -197,12 +201,13 @@ def remove_disappeared_children(base_inv, path, base_tree, existing_children,
 
     :param base_inv: Base inventory against which to generate the 
         inventory delta.
-    :param path: Path to process
+    :param path: Path to process (unicode)
     :param base_tree: Git Tree base object
     :param existing_children: Children that still exist
     :param lookup_object: Lookup a git object by its SHA1
     :return: Inventory delta, as list
     """
+    assert type(path) is unicode
     ret = []
     for name, mode, hexsha in base_tree.iteritems():
         if name in existing_children:
@@ -224,11 +229,14 @@ def import_git_tree(texts, mapping, path, name, (base_hexsha, hexsha),
     """Import a git tree object into a bzr repository.
 
     :param texts: VersionedFiles object to add to
-    :param path: Path in the tree
+    :param path: Path in the tree (str)
+    :param name: Name of the tree (str)
     :param tree: A git tree object
     :param base_inv: Base inventory against which to return inventory delta
     :return: Inventory delta for this subtree
     """
+    assert type(path) is str
+    assert type(name) is str
     if base_hexsha == hexsha and base_mode == mode:
         # If nothing has changed since the base revision, we're done
         return [], {}
@@ -242,10 +250,11 @@ def import_git_tree(texts, mapping, path, name, (base_hexsha, hexsha),
         old_path = None # Newly appeared here
     else:
         base_tree = lookup_object(base_hexsha)
-        old_path = path # Renames aren't supported yet
+        old_path = path.decode("utf-8") # Renames aren't supported yet
+    new_path = path.decode("utf-8")
     if base_tree is None or type(base_tree) is not Tree:
         ie.revision = revision_id
-        invdelta.append((old_path, path, ie.file_id, ie))
+        invdelta.append((old_path, new_path, ie.file_id, ie))
         texts.insert_record_stream([
             ChunkedContentFactory((ie.file_id, ie.revision), (), None, [])])
     # Remember for next time
@@ -266,7 +275,7 @@ def import_git_tree(texts, mapping, path, name, (base_hexsha, hexsha),
         if stat.S_ISDIR(child_mode):
             subinvdelta, grandchildmodes = import_git_tree(texts, mapping,
                 child_path, name, (child_base_hexsha, child_hexsha), base_inv,
-                file_id, revision_id, parent_invs, lookup_object, 
+                file_id, revision_id, parent_invs, lookup_object,
                 (child_base_mode, child_mode), store_updater, lookup_file_id,
                 allow_submodules=allow_submodules)
         elif S_ISGITLINK(child_mode): # submodule
@@ -289,20 +298,21 @@ def import_git_tree(texts, mapping, path, name, (base_hexsha, hexsha),
             child_modes[child_path] = child_mode
     # Remove any children that have disappeared
     if base_tree is not None and type(base_tree) is Tree:
-        invdelta.extend(remove_disappeared_children(base_inv, old_path, 
+        invdelta.extend(remove_disappeared_children(base_inv, old_path,
             base_tree, existing_children, lookup_object))
     store_updater.add_object(tree, ie, path)
     return invdelta, child_modes
 
 
 def verify_commit_reconstruction(target_git_object_retriever, lookup_object,
-    o, rev, ret_tree, parent_trees, mapping, unusual_modes):
+    o, rev, ret_tree, parent_trees, mapping, unusual_modes, verifiers):
     new_unusual_modes = mapping.export_unusual_file_modes(rev)
     if new_unusual_modes != unusual_modes:
         raise AssertionError("unusual modes don't match: %r != %r" % (
             unusual_modes, new_unusual_modes))
     # Verify that we can reconstruct the commit properly
-    rec_o = target_git_object_retriever._reconstruct_commit(rev, o.tree, True)
+    rec_o = target_git_object_retriever._reconstruct_commit(rev, o.tree, True,
+        verifiers)
     if rec_o != o:
         raise AssertionError("Reconstructed commit differs: %r != %r" % (
             rec_o, o))
@@ -335,7 +345,7 @@ def verify_commit_reconstruction(target_git_object_retriever, lookup_object,
 def import_git_commit(repo, mapping, head, lookup_object,
                       target_git_object_retriever, trees_cache):
     o = lookup_object(head)
-    rev = mapping.import_commit(o,
+    rev, roundtrip_revid, verifiers = mapping.import_commit(o,
             lambda x: target_git_object_retriever.lookup_git_sha(x)[1][0])
     # We have to do this here, since we have to walk the tree and
     # we need to make sure to import the blobs / trees with the right
@@ -350,15 +360,13 @@ def import_git_commit(repo, mapping, head, lookup_object,
         base_tree = lookup_object(o.parents[0]).tree
         base_mode = stat.S_IFDIR
     store_updater = target_git_object_retriever._get_updater(rev)
-    store_updater.add_object(o, None, None)
-    lookup_file_id = mapping.get_fileid_map(lookup_object, o.tree).lookup_file_id
+    fileid_map = mapping.get_fileid_map(lookup_object, o.tree)
     inv_delta, unusual_modes = import_git_tree(repo.texts,
-            mapping, "", u"", (base_tree, o.tree), base_inv, 
+            mapping, "", "", (base_tree, o.tree), base_inv,
             None, rev.revision_id, [p.inventory for p in parent_trees],
             lookup_object, (base_mode, stat.S_IFDIR), store_updater,
-            lookup_file_id,
+            fileid_map.lookup_file_id,
             allow_submodules=getattr(repo._format, "supports_tree_reference", False))
-    store_updater.finish()
     if unusual_modes != {}:
         for path, mode in unusual_modes.iteritems():
             warn_unusual_mode(rev.foreign_revid, path, mode)
@@ -369,15 +377,27 @@ def import_git_commit(repo, mapping, head, lookup_object,
         basis_id = NULL_REVISION
         base_inv = None
     rev.inventory_sha1, inv = repo.add_inventory_by_delta(basis_id,
-              inv_delta, rev.revision_id, rev.parent_ids,
-              base_inv)
+              inv_delta, rev.revision_id, rev.parent_ids, base_inv)
+    # Check verifiers
+    testament = StrictTestament3(rev, inv)
+    calculated_verifiers = { "testament3-sha1": testament.as_sha1() }
+    if roundtrip_revid is not None:
+        original_revid = rev.revision_id
+        rev.revision_id = roundtrip_revid
+        if calculated_verifiers != verifiers:
+            trace.mutter("Testament SHA1 %r for %r did not match %r.",
+                         calculated_verifiers["testament3-sha1"],
+                         rev.revision_id, verifiers["testament3-sha1"])
+            rev.revision_id = original_revid
+    store_updater.add_object(o, calculated_verifiers, None)
+    store_updater.finish()
     ret_tree = RevisionTree(repo, inv, rev.revision_id)
     trees_cache.add(ret_tree)
     repo.add_revision(rev.revision_id, rev)
     if "verify" in debug.debug_flags:
         verify_commit_reconstruction(target_git_object_retriever, 
             lookup_object, o, rev, ret_tree, parent_trees, mapping,
-            unusual_modes)
+            unusual_modes, verifiers)
 
 
 def import_git_objects(repo, mapping, object_iter,
@@ -409,8 +429,10 @@ def import_git_objects(repo, mapping, object_iter,
         except KeyError:
             continue
         if isinstance(o, Commit):
-            rev = mapping.import_commit(o, lambda x: None)
-            if repo.has_revision(rev.revision_id):
+            rev, roundtrip_revid, verifiers = mapping.import_commit(o,
+                lambda x: None)
+            if (repo.has_revision(rev.revision_id) or
+                (roundtrip_revid and repo.has_revision(roundtrip_revid))):
                 continue
             graph.append((o.id, o.parents))
             heads.extend([p for p in o.parents if p not in checked])
@@ -474,6 +496,17 @@ class InterGitNonGitRepository(InterGitRepository):
     """Base InterRepository that copies revisions from a Git into a non-Git
     repository."""
 
+    def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
+        """Fetch objects from a remote server.
+
+        :param determine_wants: determine_wants callback
+        :param mapping: BzrGitMapping to use
+        :param pb: Optional progress bar
+        :param limit: Maximum number of commits to import.
+        :return: Tuple with pack hint, last imported revision id and remote refs
+        """
+        raise NotImplementedError(self.fetch_objects)
+
     def fetch(self, revision_id=None, pb=None, find_ghosts=False,
               mapping=None, fetch_spec=None):
         if mapping is None:
@@ -484,18 +517,24 @@ class InterGitNonGitRepository(InterGitRepository):
             interesting_heads = fetch_spec.heads
         else:
             interesting_heads = None
-        self._refs = {}
         def determine_wants(refs):
-            self._refs = refs
             if interesting_heads is None:
-                ret = [sha for (ref, sha) in refs.iteritems() if not ref.endswith("^{}")]
+                todo = [(sha, None) for (ref, sha) in refs.iteritems() if not ref.endswith("^{}")]
             else:
-                ret = [self.source.lookup_bzr_revision_id(revid)[0] for revid in interesting_heads if revid not in (None, NULL_REVISION)]
-            return [rev for rev in ret if not self.target.has_revision(self.source.lookup_foreign_revision_id(rev))]
-        (pack_hint, _) = self.fetch_objects(determine_wants, mapping, pb)
+                todo = [(self.source.lookup_bzr_revision_id(revid)[0], revid) for revid in interesting_heads if revid not in (None, NULL_REVISION)]
+            ret = []
+            for (sha, revid) in todo:
+                if revid is None:
+                    revid = self.source.lookup_foreign_revision_id(sha)
+                if not self.target.has_revision(revid):
+                    ret.append(sha)
+            return ret
+
+        (pack_hint, _, remote_refs) = self.fetch_objects(determine_wants,
+            mapping, pb)
         if pack_hint is not None and self.target._format.pack_compresses:
             self.target.pack(hint=pack_hint)
-        return self._refs
+        return remote_refs
 
 
 _GIT_PROGRESS_RE = re.compile(r"(.*?): +(\d+)% \((\d+)/(\d+)\)")
@@ -509,6 +548,19 @@ def report_git_progress(pb, text):
         pb.update(text, 0, 0)
 
 
+class DetermineWantsRecorder(object):
+
+    def __init__(self, actual):
+        self.actual = actual
+        self.wants = []
+        self.remote_refs = {}
+
+    def __call__(self, refs):
+        self.remote_refs = refs
+        self.wants = self.actual(refs)
+        return self.wants
+
+
 class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
     """InterRepository that copies revisions from a remote Git into a non-Git
     repository."""
@@ -522,6 +574,7 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
         return set(all_revs) - all_parents
 
     def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
+        """See `InterGitNonGitRepository`."""
         def progress(text):
             report_git_progress(pb, text)
         store = BazaarObjectStore(self.target, mapping)
@@ -530,21 +583,18 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
             heads = self.get_target_heads()
             graph_walker = store.get_graph_walker(
                     [store._lookup_revision_sha1(head) for head in heads])
-            recorded_wants = []
-
-            def record_determine_wants(heads):
-                wants = determine_wants(heads)
-                recorded_wants.extend(wants)
-                return wants
+            wants_recorder = DetermineWantsRecorder(determine_wants)
 
             create_pb = None
             if pb is None:
                 create_pb = pb = ui.ui_factory.nested_progress_bar()
             try:
-                objects_iter = self.source.fetch_objects(record_determine_wants,
-                    graph_walker, store.get_raw, progress)
-                return import_git_objects(self.target, mapping,
-                    objects_iter, store, recorded_wants, pb, limit)
+                objects_iter = self.source.fetch_objects(
+                    wants_recorder, graph_walker, store.get_raw,
+                    progress)
+                (pack_hint, last_rev) = import_git_objects(self.target, mapping,
+                    objects_iter, store, wants_recorder.wants, pb, limit)
+                return (pack_hint, last_rev, wants_recorder.remote_refs)
             finally:
                 if create_pb:
                     create_pb.finished()
@@ -565,9 +615,9 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
     repository."""
 
     def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
-        """Fetch objects.
-        """
-        wants = determine_wants(self.source._git.get_refs())
+        """See `InterGitNonGitRepository`."""
+        remote_refs = self.source._git.get_refs()
+        wants = determine_wants(remote_refs)
         create_pb = None
         if pb is None:
             create_pb = pb = ui.ui_factory.nested_progress_bar()
@@ -575,9 +625,10 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
         try:
             self.target.lock_write()
             try:
-                return import_git_objects(self.target, mapping,
+                (pack_hint, last_rev) = import_git_objects(self.target, mapping,
                     self.source._git.object_store,
                     target_git_object_retriever, wants, pb, limit)
+                return (pack_hint, last_rev, remote_refs)
             finally:
                 self.target.unlock()
         finally:
@@ -602,8 +653,9 @@ class InterGitGitRepository(InterGitRepository):
         graphwalker = self.target._git.get_graph_walker()
         if (isinstance(self.source, LocalGitRepository) and
             isinstance(self.target, LocalGitRepository)):
-            return self.source._git.fetch(self.target._git, determine_wants,
+            refs = self.source._git.fetch(self.target._git, determine_wants,
                 progress)
+            return (None, None, refs)
         elif (isinstance(self.source, LocalGitRepository) and
               isinstance(self.target, RemoteGitRepository)):
             raise NotImplementedError
@@ -611,10 +663,10 @@ class InterGitGitRepository(InterGitRepository):
               isinstance(self.target, LocalGitRepository)):
             f, commit = self.target._git.object_store.add_thin_pack()
             try:
-                refs = self.source._git.fetch_pack(determine_wants,
-                    graphwalker, f.write, progress)
+                refs = self.source.bzrdir.root_transport.fetch_pack(
+                    determine_wants, graphwalker, f.write, progress)
                 commit()
-                return refs
+                return (None, None, refs)
             except:
                 f.close()
                 raise
@@ -636,8 +688,7 @@ class InterGitGitRepository(InterGitRepository):
             determine_wants = r.object_store.determine_wants_all
         else:
             determine_wants = lambda x: [y for y in args if not y in r.object_store]
-        return self.fetch_objects(determine_wants, mapping)[0]
-
+        self.fetch_objects(determine_wants, mapping)
 
     @staticmethod
     def is_compatible(source, target):
