@@ -104,7 +104,7 @@ class BzrGitMapping(foreign.VcsMapping):
         super(BzrGitMapping, self).__init__(foreign_git)
 
     def __eq__(self, other):
-        return (type(self) == type(other) and 
+        return (type(self) == type(other) and
                 self.revid_prefix == other.revid_prefix)
 
     @classmethod
@@ -185,7 +185,7 @@ class BzrGitMapping(foreign.VcsMapping):
 
     def _extract_git_svn_metadata(self, rev, message):
         lines = message.split("\n")
-        if not (lines[-1] == "" and lines[-2].startswith("git-svn-id:")):
+        if not (lines[-1] == "" and len(lines) >= 2 and lines[-2].startswith("git-svn-id:")):
             return message
         git_svn_id = lines[-2].split(": ", 1)[1]
         rev.properties['git-svn-id'] = git_svn_id
@@ -227,12 +227,15 @@ class BzrGitMapping(foreign.VcsMapping):
         b.set_raw_chunks(serialize_fileid_map(fileid_map))
         return b
 
-    def export_commit(self, rev, tree_sha, parent_lookup, roundtrip):
+    def export_commit(self, rev, tree_sha, parent_lookup, roundtrip,
+                      verifiers):
         """Turn a Bazaar revision in to a Git commit
 
         :param tree_sha: Tree sha for the commit
         :param parent_lookup: Function for looking up the GIT sha equiv of a
             bzr revision
+        :param roundtrip: Whether to store roundtripping information.
+        :param verifiers: Verifiers info
         :return dulwich.objects.Commit represent the revision:
         """
         from dulwich.objects import Commit
@@ -240,6 +243,7 @@ class BzrGitMapping(foreign.VcsMapping):
         commit.tree = tree_sha
         if roundtrip:
             metadata = BzrGitRevisionMetadata()
+            metadata.verifiers = verifiers
         else:
             metadata = None
         parents = []
@@ -307,8 +311,8 @@ class BzrGitMapping(foreign.VcsMapping):
     def import_commit(self, commit, lookup_parent_revid):
         """Convert a git commit to a bzr revision.
 
-        :return: a `bzrlib.revision.Revision` object and a 
-            dictionary of path -> file ids
+        :return: a `bzrlib.revision.Revision` object, foreign revid and a
+            testament sha1
         """
         if commit is None:
             raise AssertionError("Commit object can't be None")
@@ -347,12 +351,15 @@ class BzrGitMapping(foreign.VcsMapping):
         rev.timezone = commit.commit_timezone
         if rev.git_metadata is not None:
             md = rev.git_metadata
-            if md.revision_id:
-                rev.revision_id = md.revision_id
+            roundtrip_revid = md.revision_id
             if md.explicit_parent_ids:
                 rev.parent_ids = md.explicit_parent_ids
             rev.properties.update(md.properties)
-        return rev
+            verifiers = md.verifiers
+        else:
+            roundtrip_revid = None
+            verifiers = {}
+        return rev, roundtrip_revid, verifiers
 
     def get_fileid_map(self, lookup_object, tree_sha):
         """Obtain a fileid map for a particular tree.
@@ -400,9 +407,9 @@ class BzrGitMappingExperimental(BzrGitMappingv1):
         return ret
 
     def import_commit(self, commit, lookup_parent_revid):
-        rev, file_ids = super(BzrGitMappingExperimental, self).import_commit(commit, lookup_parent_revid)
+        rev, roundtrip_revid, verifiers = super(BzrGitMappingExperimental, self).import_commit(commit, lookup_parent_revid)
         rev.properties['converted_revision'] = "git %s\n" % commit.id
-        return rev, file_ids
+        return rev, roundtrip_revid, verifiers
 
 
 class GitMappingRegistry(VcsMappingRegistry):
@@ -574,6 +581,7 @@ class GitFileIdMap(object):
         self.mapping = mapping
 
     def lookup_file_id(self, path):
+        assert type(path) is str
         try:
             file_id = self.file_ids[path]
         except KeyError:
@@ -587,6 +595,9 @@ class GitFileIdMap(object):
             for k, v in self.file_ids.iteritems():
                 self.paths[v] = k
         try:
-            return self.paths[file_id]
+            path = self.paths[file_id]
         except KeyError:
             return self.mapping.parse_file_id(file_id)
+        else:
+            assert type(path) is str
+            return path
