@@ -18,6 +18,7 @@
 
 from dulwich.objects import (
     Blob,
+    Commit,
     Tree,
     sha_to_hex,
     )
@@ -35,6 +36,9 @@ from bzrlib import (
 from bzrlib.revision import (
     NULL_REVISION,
     )
+from bzrlib.testament import(
+    StrictTestament3,
+    )
 
 from bzrlib.plugins.git.mapping import (
     default_mapping,
@@ -73,16 +77,26 @@ class LRUTreeCache(object):
 
     def revision_tree(self, revid):
         try:
-            return self._cache[revid]
+            tree = self._cache[revid]
         except KeyError:
             tree = self.repository.revision_tree(revid)
             self.add(tree)
-            return tree
+        assert tree.get_revision_id() == tree.inventory.revision_id
+        return tree
 
     def iter_revision_trees(self, revids):
-        trees = dict([(k, self._cache.get(k)) for k in revids]) 
-        for tree in self.repository.revision_trees(
-                [r for r, v in trees.iteritems() if v is None]):
+        trees = {}
+        todo = []
+        for revid in revids:
+            try:
+                tree = self._cache[revid]
+            except KeyError:
+                todo.append(revid)
+            else:
+                assert tree.get_revision_id() == revid
+                assert tree.inventory.revision_id == revid
+                trees[revid] = tree
+        for tree in self.repository.revision_trees(todo):
             trees[tree.get_revision_id()] = tree
             self.add(tree)
         return (trees[r] for r in revids)
@@ -94,7 +108,7 @@ class LRUTreeCache(object):
         self._cache.add(tree.get_revision_id(), tree)
 
 
-def _find_missing_bzr_revids(get_parent_map, want, have):
+def _find_missing_bzr_revids(graph, want, have):
     """Find the revisions that have to be pushed.
 
     :param get_parent_map: Function that returns the parents for a sequence
@@ -103,19 +117,9 @@ def _find_missing_bzr_revids(get_parent_map, want, have):
     :param have: Revisions the target already has
     :return: Set of revisions to fetch
     """
-    pending = want - have
-    processed = set()
     todo = set()
-    while pending:
-        processed.update(pending)
-        next_map = get_parent_map(pending)
-        next_pending = set()
-        for item in next_map.iteritems():
-            if item[0] in have:
-                continue
-            todo.add(item[0])
-            next_pending.update(p for p in item[1] if p not in processed)
-        pending = next_pending
+    for rev in want:
+        todo.update(graph.find_unique_ancestors(rev, have))
     if NULL_REVISION in todo:
         todo.remove(NULL_REVISION)
     return todo
@@ -210,7 +214,9 @@ def _tree_to_objects(tree, parent_trees, idmap, unusual_modes,
             new_trees[posixpath.dirname(path[1])] = parent[1]
         elif kind[1] not in (None, "directory"):
             raise AssertionError(kind[1])
-        if path[0] is not None and parent[0] in tree.inventory and tree.inventory[parent[0]].kind == "directory":
+        if (path[0] not in (None, "") and
+            parent[0] in tree.inventory and
+            tree.inventory[parent[0]].kind == "directory"):
             # Removal
             new_trees[posixpath.dirname(path[0])] = parent[0]
     
@@ -225,7 +231,7 @@ def _tree_to_objects(tree, parent_trees, idmap, unusual_modes,
     for path in unusual_modes:
         parent_path = posixpath.dirname(path)
         new_trees[parent_path] = tree.path2id(parent_path)
-    
+
     trees = {}
     while new_trees:
         items = new_trees.items()
@@ -326,14 +332,22 @@ class BazaarObjectStore(BaseObjectStore):
         self._update_sha_map()
         return iter(self._cache.idmap.sha1s())
 
-    def _reconstruct_commit(self, rev, tree_sha, roundtrip):
+    def _reconstruct_commit(self, rev, tree_sha, roundtrip, verifiers):
+        """Reconstruct a Commit object.
+
+        :param rev: Revision object
+        :param tree_sha: SHA1 of the root tree object
+        :param roundtrip: Whether or not to roundtrip bzr metadata
+        :param verifiers: Verifiers for the commits
+        :return: Commit object
+        """
         def parent_lookup(revid):
             try:
                 return self._lookup_revision_sha1(revid)
             except errors.NoSuchRevision:
                 return None
         return self.mapping.export_commit(rev, tree_sha, parent_lookup,
-            roundtrip)
+            roundtrip, verifiers)
 
     def _create_fileid_map_blob(self, inv):
         # FIXME: This can probably be a lot more efficient, 
@@ -378,8 +392,13 @@ class BazaarObjectStore(BaseObjectStore):
                 root_tree[self.mapping.BZR_FILE_IDS_FILE] = ((stat.S_IFREG | 0644), b.id)
                 yield self.mapping.BZR_FILE_IDS_FILE, b, None
         yield "", root_tree, root_ie
+        if roundtrip:
+            testament3 = StrictTestament3(rev, tree.inventory)
+            verifiers = { "testament3-sha1": testament3.as_sha1() }
+        else:
+            verifiers = {}
         commit_obj = self._reconstruct_commit(rev, root_tree.id,
-            roundtrip=roundtrip)
+            roundtrip=roundtrip, verifiers=verifiers)
         try:
             foreign_revid, mapping = mapping_registry.parse_revision_id(
                 rev.revision_id)
@@ -398,6 +417,9 @@ class BazaarObjectStore(BaseObjectStore):
         updater = self._get_updater(rev)
         for path, obj, ie in self._revision_to_objects(rev, tree,
             roundtrip=True):
+            if isinstance(obj, Commit):
+                testament3 = StrictTestament3(rev, tree.inventory)
+                ie = { "testament3-sha1": testament3.as_sha1() }
             updater.add_object(obj, ie, path)
         commit_obj = updater.finish()
         return commit_obj.id
@@ -515,7 +537,7 @@ class BazaarObjectStore(BaseObjectStore):
         ret = {}
         for sha in shas:
             if sha == ZERO_SHA:
-                ret[sha] = NULL_REVISION
+                ret[sha] = ("commit", (NULL_REVISION, None, {}))
                 continue
             try:
                 ret[sha] = self._cache.idmap.lookup_git_sha(sha)
@@ -543,14 +565,15 @@ class BazaarObjectStore(BaseObjectStore):
         (type, type_data) = self.lookup_git_sha(sha)
         # convert object to git object
         if type == "commit":
-            (revid, tree_sha) = type_data
+            (revid, tree_sha, verifiers) = type_data
             try:
                 rev = self.repository.get_revision(revid)
             except errors.NoSuchRevision:
                 trace.mutter('entry for %s %s in shamap: %r, but not found in '
                              'repository', type, sha, type_data)
                 raise KeyError(sha)
-            commit = self._reconstruct_commit(rev, tree_sha, roundtrip=True)
+            commit = self._reconstruct_commit(rev, tree_sha, roundtrip=True,
+                verifiers=verifiers)
             _check_expected_sha(sha, commit)
             return commit
         elif type == "blob":
@@ -589,7 +612,7 @@ class BazaarObjectStore(BaseObjectStore):
         ret = self.lookup_git_shas(have + want)
         for commit_sha in have:
             try:
-                (type, (revid, tree_sha)) = ret[commit_sha]
+                (type, (revid, tree_sha, verifiers)) = ret[commit_sha]
             except KeyError:
                 pass
             else:
@@ -600,22 +623,25 @@ class BazaarObjectStore(BaseObjectStore):
             if commit_sha in have:
                 continue
             try:
-                (type, (revid, tree_sha)) = ret[commit_sha]
+                (type, (revid, tree_sha, verifiers)) = ret[commit_sha]
             except KeyError:
                 pass
             else:
                 assert type == "commit"
                 pending.add(revid)
 
-        todo = _find_missing_bzr_revids(self.repository.get_parent_map, 
-                                        pending, processed)
+        graph = self.repository.get_graph()
+        todo = _find_missing_bzr_revids(graph, pending, processed)
         trace.mutter('sending revisions %r', todo)
         ret = []
         pb = ui.ui_factory.nested_progress_bar()
         try:
             for i, revid in enumerate(todo):
                 pb.update("generating git objects", i, len(todo))
-                rev = self.repository.get_revision(revid)
+                try:
+                    rev = self.repository.get_revision(revid)
+                except errors.NoSuchRevision:
+                    continue
                 tree = self.tree_cache.revision_tree(revid)
                 for path, obj, ie in self._revision_to_objects(rev, tree,
                     roundtrip=not lossy):
