@@ -16,16 +16,19 @@
 
 """The basic test suite for bzr-git."""
 
-import subprocess
+from cStringIO import StringIO
+
 import time
 
 from bzrlib import (
-    osutils,
+    errors as bzr_errors,
     tests,
-    trace,
     )
 from bzrlib.plugins.git import (
-    errors,
+    import_dulwich,
+    )
+from fastimport import (
+    commands,
     )
 
 TestCase = tests.TestCase
@@ -33,92 +36,48 @@ TestCaseInTempDir = tests.TestCaseInTempDir
 TestCaseWithTransport = tests.TestCaseWithTransport
 TestCaseWithMemoryTransport = tests.TestCaseWithMemoryTransport
 
-class _GitCommandFeature(tests.Feature):
+class _DulwichFeature(tests.Feature):
 
     def _probe(self):
         try:
-            p = subprocess.Popen(['git', '--version'], stdout=subprocess.PIPE)
-        except IOError:
+            import_dulwich()
+        except bzr_errors.DependencyNotPresent:
             return False
-        out, err = p.communicate()
-        trace.mutter('Using: %s', out.rstrip('\n'))
         return True
 
     def feature_name(self):
-        return 'git'
-
-GitCommandFeature = _GitCommandFeature()
+        return 'dulwich'
 
 
-def run_git(*args):
-    cmd = ['git'] + list(args)
-    p = subprocess.Popen(cmd,
-                         stdout=subprocess.PIPE,
-                         stderr=subprocess.PIPE)
-    out, err = p.communicate()
-    if p.returncode != 0:
-        raise AssertionError('Bad return code: %d for %s:\n%s'
-                             % (p.returncode, ' '.join(cmd), err))
-    return out
+DulwichFeature = _DulwichFeature()
 
 
 class GitBranchBuilder(object):
 
     def __init__(self, stream=None):
         self.commit_info = []
-        self.stream = stream
-        self._process = None
+        self.orig_stream = stream
+        if stream is None:
+            self.stream = StringIO()
+        else:
+            self.stream = stream
         self._counter = 0
         self._branch = 'refs/heads/master'
-        if stream is None:
-            # Write the marks file into the git sandbox.
-            self._marks_file_name = osutils.abspath('marks')
-            self._process = subprocess.Popen(
-                ['git', 'fast-import', '--quiet',
-                 # GIT doesn't support '--export-marks foo'
-                 # it only supports '--export-marks=foo'
-                 # And gives a 'unknown option' otherwise.
-                 '--export-marks=' + self._marks_file_name,
-                ],
-                stdout=subprocess.PIPE,
-                stderr=subprocess.PIPE,
-                stdin=subprocess.PIPE,
-                )
-            self.stream = self._process.stdin
-        else:
-            self._process = None
 
     def set_branch(self, branch):
         """Set the branch we are committing."""
         self._branch = branch
 
     def _write(self, text):
-        try:
-            self.stream.write(text)
-        except IOError, e:
-            if self._process is None:
-                raise
-            raise errors.GitCommandError(self._process.returncode,
-                                         'git fast-import',
-                                         self._process.stderr.read())
+        self.stream.write(text)
 
     def _writelines(self, lines):
-        try:
-            self.stream.writelines(lines)
-        except IOError, e:
-            if self._process is None:
-                raise
-            raise errors.GitCommandError(self._process.returncode,
-                                         'git fast-import',
-                                         self._process.stderr.read())
+        self.stream.writelines(lines)
 
     def _create_blob(self, content):
         self._counter += 1
-        self._write('blob\n')
-        self._write('mark :%d\n' % (self._counter,))
-        self._write('data %d\n' % (len(content),))
-        self._write(content)
-        self._write('\n')
+        blob = commands.BlobCommand(str(self._counter), content)
+        self._write(str(blob)+"\n")
         return self._counter
 
     def set_symlink(self, path, content):
@@ -177,11 +136,11 @@ class GitBranchBuilder(object):
             commit.
         """
         self._counter += 1
-        mark = self._counter
+        mark = str(self._counter)
         if timestamp is None:
             timestamp = int(time.time())
         self._write('commit %s\n' % (self._branch,))
-        self._write('mark :%d\n' % (mark,))
+        self._write('mark :%s\n' % (mark,))
         self._write('committer %s %s %s\n'
                     % (committer, timestamp, timezone))
         message = message.encode('UTF-8')
@@ -189,10 +148,10 @@ class GitBranchBuilder(object):
         self._write(message)
         self._write('\n')
         if base is not None:
-            self._write('from :%d\n' % (base,))
+            self._write('from :%s\n' % (base,))
         if merge is not None:
             for m in merge:
-                self._write('merge :%d\n' % (m,))
+                self._write('merge :%s\n' % (m,))
         self._writelines(self.commit_info)
         self._write('\n')
         self.commit_info = []
@@ -208,32 +167,24 @@ class GitBranchBuilder(object):
             ref = self._branch
         self._write('reset %s\n' % (ref,))
         if mark is not None:
-            self._write('from :%d\n' % mark)
+            self._write('from :%s\n' % mark)
         self._write('\n')
 
     def finish(self):
         """We are finished building, close the stream, get the id mapping"""
-        self.stream.close()
-        if self._process is None:
-            return {}
-        if self._process.wait() != 0:
-            raise errors.GitCommandError(self._process.returncode,
-                                         'git fast-import',
-                                         self._process.stderr.read())
-        marks_file = open(self._marks_file_name)
-        mapping = {}
-        for line in marks_file:
-            mark, shasum = line.split()
-            assert mark.startswith(':')
-            mapping[int(mark[1:])] = shasum
-        marks_file.close()
-        return mapping
+        self.stream.seek(0)
+        if self.orig_stream is None:
+            from dulwich.repo import Repo
+            r = Repo(".")
+            from dulwich.fastexport import GitImportProcessor
+            importer = GitImportProcessor(r)
+            return importer.import_stream(self.stream)
 
 
 def test_suite():
-    loader = tests.TestLoader()
+    loader = tests.TestUtil.TestLoader()
 
-    suite = tests.TestSuite()
+    suite = tests.TestUtil.TestSuite()
 
     testmod_names = [
         'test_blackbox',
